/*
**   IRC - Internet Relay Chat, source/ircd.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"

RCSID("$Id: ircd.c,v 1.34 1999/02/23 07:57:20 cbehrens Exp $");


typedef struct _ircd_data	IRCD_DATA;

struct _ircd_data
{
	int				Maxconnections;
	int				Maxfdlists;

	/* LEAVE THE ABOVE 2 FIRST AND NEVER CHANGE !!! */

	char			*Conffile;
	int				ircd_pagesize;
	FDLIST			*Fdlist_main;
	FDLIST			*Fdlist_clients;
	int				*Fdlist_lasttime;
	IRCD_CLIENT		**Fds;				/* for easy fd -> client lookup */
	IRCD_CLIENT		*Locals;
	IRCD_CLIENT		*Clients;
	IRCD_CLIENT		*Opers;
	IRCD_CLIENT		*Servers;
	IRCD_CLIENT		*Flooders;
	IRCD_CHANNEL	*Channels;
	GLOBALSTATS		*Globalinfo;
	IRCD_CLIENT		*Rejectholds;
	IRCD_MOTD		*Motd;
	int				Sendbuffersize;
	int				Recvbuffersize;

	unsigned int	*Sentalong;
	unsigned int	Sentalong_marker;
	unsigned int	*Sentalong_end;

	int				Ircd_resfd;
	IRCD_CLIENT		*Ircd_resclient;

	int				Ircd_zero_fd;

	IRCDCONF		Ircd_conf;

	IRCD_CLIENT		*meptr;

	int				queue_size;
	int				queue_malloced_size;

	void			*res_requests;
	void			*res_lastrequest;
	char			**NumericArr;
	int				NumericArrSize;

	void			*whowasarr;
	int				whowas_size;
	void			**whowashash;
	int				whowas_hash_size;
	int				whowas_next;
	void			*Vars;
	void			**client_hash;
	int				client_hash_size;
	void			**channel_hash;
	int				channel_hash_size;
	void			***iphash;
	int				iphash_size;
	void			*clones;
	void			*throttles;
};

int				Maxconnections;
int				Maxfdlists;
char			*Conffile;
int				ircd_pagesize;
FDLIST			*Fdlist_main;
static FDLIST	*Fdlist_clients;
static int		*Fdlist_lasttime;
IRCD_CLIENT		**Fds;				/* for easy fd -> client lookup */
IRCD_CLIENT		*Locals = NULL;			/* local clients */
IRCD_CLIENT		*Clients = NULL;
IRCD_CLIENT		*Opers = NULL;
IRCD_CLIENT		*Servers = NULL;
IRCD_CLIENT		*Flooders = NULL;
IRCD_CHANNEL	*Channels = NULL;
GLOBALSTATS		*Globalinfo;
IRCD_CLIENT		*Rejectholds = NULL;
IRCD_MOTD		*Motd = NULL;
int				Sendbuffersize = 0;
int				Recvbuffersize = 0;
unsigned int	*Sentalong;
unsigned int	Sentalong_marker;
unsigned int	*Sentalong_end;
IRCD_CLIENT		*meptr;
int				Ircd_zero_fd = -1;
IRCDCONF		_ircd_conf;

int				Now;
int				DoRehash = 0;
char			*RehashWho = NULL;
int				Rehashed = 0;
int				Autoconncheck = 0;
int				DoReload = 0;


static int		ircdcore_argc;
static char		**ircdcore_argv;
static char		*ircdcore_versionstring;
static int		*ircdcore_data_format_version;
static char		**ircdcore_data;
static int		*ircdcore_data_size;
static FDLIST_POLLFD *pfdarray = NULL;

char	_ircdso_versionstring[] = VERSION;
int		_ircdso_version = 1;

static IRCD_CLIENT	*Ircd_resclient;
static int			Ircd_resfd = -1;

static RETSIGTYPE _ircd_sighup_handler(int sig);
static void res_read(IRCD_CLIENT **cptr, SERVLINK **slptr);
static int _ircd_pollfds(FDLIST *fdlist, int secs);
static void flood_check(void);
static void ping_check(void);
static void kline_check(void);
static void rejecthold_check(void);
static int autoconn_check(void);
static int _ircdso_firsttime(int argc, char **argv);
static int _ircdso_init(int firsttime);
static int _ircdso_deinit(void);

static RETSIGTYPE _ircd_sighup_handler(int sig)
{
	DoRehash = 1;
	if (!RehashWho)
		RehashWho = MyStrdup("SIGHUP");
	signal(sig, _ircd_sighup_handler);
}

static RETSIGTYPE _ircd_sigusr1_handler(int sig)
{
	DoReload = 1;
	signal(sig, _ircd_sigusr1_handler);
}

static void res_read(IRCD_CLIENT **clptr, SERVLINK **slptr)
{
	MYRES_STATUS	*sptr;
	MYRES_HOSTENT	*heptr;
	IRCD_CLIENT		*cptr = NULL;
	SERVLINK		*slink = NULL;

	*clptr = NULL;
	*slptr = NULL;
	sptr = myres_get_answer();
	if (!sptr)
		return;
	if (sptr->ctype == MYRES_TYPE_CLIENT)
	{
		cptr = (IRCD_CLIENT *)sptr->cptr;
		cptr->res_stuff->res_status.rptr = NULL;
	}
	else if (sptr->ctype == MYRES_TYPE_SERVLINK)
	{
		slink = (SERVLINK *)sptr->cptr;
		slink->res_status.rptr = NULL;
	}
	else
		return;
	if (!sptr->result) /* FAILED */
	{
		switch(sptr->ctype)
		{
			case MYRES_TYPE_CLIENT:
				cptr->flags &= ~CLIENT_FLAGS_DOINGDNS;
				*clptr = cptr;
				break;
			case MYRES_TYPE_SERVLINK:
				send_to_opers("Lookup of %s for \"link %s\" conf entry failed",
					slink->hostname, slink->servername);
				slink->flags &= ~SERVLINK_FLAGS_DOINGDNS;
				*slptr = slink;
				break;
		}
		return;
	}
	if (sptr->ctype == MYRES_TYPE_CLIENT)
	{
		if (sptr->type == MYRES_TYPE_PTR)
		{
			myres_hostent_move(&(cptr->res_stuff->he_reverse), &(sptr->he));
			sptr->type = 0;
			heptr = myres_gethostbyname(cptr->res_stuff->he_reverse.h.h_name, sptr);
			if (heptr)
			{
				myres_hostent_move(&(cptr->res_stuff->he_forward), &(sptr->he));
				MyFree(heptr);
			}
			if (!heptr)
			{
				*clptr = cptr;
				return;
			}
		}
		else
			myres_hostent_move(&(cptr->res_stuff->he_forward), &(sptr->he));
		myres_status_deinit(&(cptr->res_stuff->res_status));

		/* check reverse vs forward */
		if (sptr->result)
		{
			ircd_uint32_t	ipaddr = htonl(cptr->rem_ipaddr);
			char			**addrs = cptr->res_stuff->he_forward.h.h_addr_list;

			for(;*addrs;addrs++)
				if (!BCMP(*addrs, (char *)&ipaddr, sizeof(struct in_addr)))
					break;
			if (!*addrs)
			{
				send_to_flag(CLIENT_NFLAGS_IPMISMATCHES, NULL,
					"IP Mismatch: %s != %s[%s]",
					cptr->hostname,
					cptr->res_stuff->he_forward.h.h_name ?
						cptr->res_stuff->he_forward.h.h_name : "null",
					cptr->res_stuff->he_forward.h.h_addr_list[0] ? 
					inet_ntoa(*(struct in_addr *)
							cptr->res_stuff->he_forward.h.h_addr_list[0]) : "null");
			}
			else
			{
				if (strlen(cptr->res_stuff->he_forward.h.h_name) <=
								sizeof(cptr->hostname)-1)
					strcpy(cptr->hostname, cptr->res_stuff->he_forward.h.h_name);
			}
		}
		cptr->flags &= ~CLIENT_FLAGS_DOINGDNS;
		*clptr = cptr;
		return;
	}
	/* MYRES_TYPE_SERVLINK */
	if (sptr->type != MYRES_TYPE_A)
	{
		myres_status_deinit(&(slink->res_status));
		slink->flags &= ~SERVLINK_FLAGS_DOINGDNS;
		*slptr = slink;
		return;
	}
	myres_hostent_move(&(slink->he_forward), &(sptr->he));
	myres_status_deinit(&(slink->res_status));
	slink->flags &= ~SERVLINK_FLAGS_DOINGDNS;
	if (slink->ipaddr == -1)
	{
        if (slink->he_forward.h.h_addr_list[0])
        {
            BCOPY(slink->he_forward.h.h_addr_list[0], &slink->ipaddr, sizeof(slink->ipaddr));
            slink->ipaddr = htonl(slink->ipaddr);
        }
	}
	*slptr = slink;
	return;
}

static int _ircd_pollfds(FDLIST *fdlist, int secs)
{
#ifndef HAVE_POLL
	fd_set			readfds;
	fd_set			writefds;
	struct timeval	tv;
#endif
	int				lowestfd;
	register int	i;
	int				numfds;
	int				ret;
	int				nfds;
	static char		readbuf[1024*64];
	static int		readbufsize = 1024*64;
	IRCD_CLIENT		*cptr;
	char			exitbuf[512];

	if (!fdlist->pfd_num)
		return 0;

#ifdef HAVE_POLL
	BCOPY(fdlist->pfd_array, pfdarray, sizeof(FDLIST_POLLFD)*fdlist->pfd_num); 
	numfds = fdlist->pfd_num;
	lowestfd = 0;
#else
	bcopy(&fdlist->read_fds, &readfds, sizeof(readfds));
	bcopy(&fdlist->write_fds, &writefds, sizeof(writefds));
	numfds = Maxconnections;
	lowestfd = fdlist->lowest_fd;
#endif
	for(;;) /* fake loop */
	{
#ifdef HAVE_POLL
		nfds = poll(pfdarray, numfds, secs*1000);
#else
		tv.tv_sec = secs;
		tv.tv_usec = 0;
		nfds = select(numfds+1, &readfds, &writefds, NULL, &tv);
#endif
		if (nfds < 0)
		{
			Now = time(NULL);
			return nfds;
		}
		Now = time(NULL);
		for(i=lowestfd;nfds && i<numfds;i++)
		{
#ifdef HAVE_POLL
			if (!pfdarray[i].revents)
#else
			if (!FD_ISSET(i, &readfds) && !FD_ISSET(i, &writefds))
#endif
				continue;
			nfds--;
#ifdef HAVE_POLL
			if ((cptr = GET_FDS(pfdarray[i].fd)) == NULL)
#else
			if ((cptr = GET_FDS(i)) == NULL)
#endif
				continue;
			cptr->lasttime = Now;
			if (IsRes(cptr))
			{
				SERVLINK	*slink;

				res_read(&cptr, &slink);
				if (cptr)
				{
					if (!(cptr->flags &
								(CLIENT_FLAGS_DOINGAUTH|CLIENT_FLAGS_DOINGDNS)))
					{
						if (ircd_parse(cptr) < 0)
							continue;
					}
				}
				else if (slink)
				{

				}
				continue;
			}
#ifdef HAVE_POLL
			if (pfdarray[i].revents & (POLLERR|POLLHUP|POLLNVAL))
				pfdarray[i].revents |= POLLIN; /* will close() upon read()*/
			if (pfdarray[i].revents & POLLIN)
#else
			if (FD_ISSET(i, &readfds))
#endif
			{
				if (IsListener(cptr))
				{
					struct sockaddr_in	sin;
					int					sinsize = sizeof(sin);

					ret = accept(cptr->fdlistinfo.fd,
								(struct sockaddr *)&sin, &sinsize);
					if (ret < 0)
						continue;
					if (ret >= Maxconnections)
						close(ret);
					else
					{
						int ipaddr = ntohl(sin.sin_addr.s_addr);

						if (iphash_handle_accept(ipaddr))
						{
							int num;

							if ((num=(int)vars_value_get_bynum(IRCD_VAR_REJECT_HOLD)) > 0)
							{
								if (client_add_rejecthold(ret, &sin) >= 0)
									continue;
							}
							close(ret);
							continue;
						}

						if (!access_match_ipaddr(&(_ircd_conf.deny),
							NULL, NULL, ipaddr, 0, NULL))
						{
							DEBUG(DEBUG_ALL, ("denying accept from %s\n",
								inet_ntoa(sin.sin_addr)));
							close(ret);
						}
						else if (client_add_local(cptr, ret, &sin) < 0)
							close(ret);
					}
					continue;
				}
				ret = read(cptr->fdlistinfo.fd, readbuf, readbufsize-1);
				if ((ret < 0) && 
						((errno == EAGAIN) || (errno == EINTR)
							|| (errno == EWOULDBLOCK)))
					continue;
				if (ret <= 0)
				{
					IRCD_CLIENT *cptrkeep = cptr->auth_ptr;

					if (ret == 0)
						sprintf(exitbuf, "EOF of client");
					else
						sprintf(exitbuf, "Read error %i: %s", errno,
							(errno == 0) ? "Undefined error" : 
									support_strerror(errno));
					if (!IsAuth(cptr))
					{
						client_del(meptr, cptr, exitbuf);
						continue;
					}
					client_del(meptr, cptr, exitbuf);
					cptr = cptrkeep; /* fall through to the ircd_parse */
				}
				else
					readbuf[ret] = '\0';
				if (IsRejecthold(cptr))
					continue;
				if (ret > 0 && (queue_data_add(&(cptr->queue_in), readbuf, ret) < 0))
				{
					IRCD_CLIENT *cptrkeep = cptr->auth_ptr;

					if (!IsAuth(cptr))
					{
						client_del(meptr, cptr, "Couldn't add to RecvQ");
						continue;
					}
					client_del(meptr, cptr, "Couldn't add to RecvQ");
					cptr = cptrkeep; /* fall through to the ircd_parse */
				}
				else if (ret > 0)
				{
					cptr->recv_bytes += ret;
					if (cptr->recv_bytes >= 1024)
					{
						cptr->recv_kbytes += cptr->recv_bytes / 1024;
						cptr->recv_bytes %= 1024;
					}
					if (IsClient(cptr))
						Globalinfo->bw_c_incoming_bytes += ret;
					else if (IsServer(cptr))
						Globalinfo->bw_s_incoming_bytes += ret;
				}
				if (IsAuth(cptr))
				{
					char	*buf = NULL;
					int		len = 0;
					short	remp, locp;
					char	ruser[30];
					IRCD_CLIENT	*cptrkeep;
					char	tuser[128];
					char	*blah = tuser;

					if (queue_data_get(&(cptr->queue_in), &buf, &len) < 0)
						continue;

					if (len <= sizeof(tuser)-1)
					{
						strncpy(blah, buf, len);
					}
					else
					{
						blah = (char *)MyMalloc(len+1);
						strncpy(blah, buf, len);
					}
					blah[len] = '\0';

					strncpy(tuser, buf, sizeof(tuser)-1);
					if (len < sizeof(tuser))
						tuser[len] = '\0';
					else
						tuser[sizeof(tuser)-1] = '\0';
					if ((len > 10) && sscanf(tuser, "%hd , %hd : USERID : %*[^:]: %10s",
						&remp, &locp, ruser) == 3)
					{
						char *s = strrchr(tuser, ':') + 1;
						char *t;

						len = 0;
						for(t=ruser;*s && (len<(sizeof(ruser)-1));s++)
							if (!isspace(*s) && (*s != ':') && (*s != '@'))
							{
								len++;
								*t++ = *s;
							}
						*t = '\0';
						strncpy(cptr->auth_ptr->authusername, ruser,
								sizeof(cptr->auth_ptr->authusername)-1);
						strncpy(cptr->auth_ptr->username, ruser,
								sizeof(cptr->auth_ptr->username)-1);
						if (*ruser)
							cptr->auth_ptr->flags |= CLIENT_FLAGS_GOTAUTH;
					}
					else
					{
						if (!strchr(buf, '\n') && !strchr(buf, '\r'))
						{
							if (blah != tuser)
								MyFree(blah);
							continue;
						}
					}
					cptrkeep = cptr->auth_ptr;
					client_del(cptr, cptr, NULL);
					cptr = cptrkeep; /* fall through to the ircd_parse */
					if (blah != tuser)
						MyFree(blah);
				}
				if ((IsClient(cptr) && (cptr->queue_in.queue_size > 8192)) ||
					(!IsServer(cptr) && (cptr->queue_in.queue_size > 32768)))
				{
					IRCD_CLIENT *cptrkeep = cptr->auth_ptr;

					if (!IsAuth(cptr))
					{
						client_del(meptr, cptr, "Excess Flood");
						continue;
					}
					client_del(meptr, cptr, "Excess Flood");
					cptr = cptrkeep; /* fall through to the ircd_parse */
				}
				if (!(cptr->flags &
							(CLIENT_FLAGS_DOINGAUTH|CLIENT_FLAGS_DOINGDNS)))
				{
					if (IsServer(cptr) &&
								(cptr->flags & CLIENT_FLAGS_DOINGBURST))
					{
						if ((Now - cptr->lastparse) > 4)
						{
							if (ircd_parse(cptr) < 0)
								continue;
							else
								cptr->lastparse = Now;
						}
					}
					else if (ircd_parse(cptr) < 0)
						continue;
					else
						cptr->lastparse = Now;
				}
			}
#ifdef HAVE_POLL
			if (pfdarray[i].revents & POLLOUT)
#else
			if (FD_ISSET(i, &writefds))
#endif
			{
				if (cptr->flags & CLIENT_FLAGS_DOINGCONNECT)
				{
					cptr->flags &= ~CLIENT_FLAGS_DOINGCONNECT;
					/* connected to wherever */
					if (IsAuth(cptr))
					{
						char buffer[64];
						int	len;

						sprintf(buffer, "%u , %u\r\n",
							(unsigned int)cptr->auth_ptr->rem_port,
							(unsigned int)cptr->auth_ptr->loc_port);
						len = strlen(buffer);
						if (write(cptr->fdlistinfo.fd, buffer, len) != len)
						{
							IRCD_CLIENT *cptrkeep = cptr->auth_ptr;
							client_del(meptr, cptr, "write error");
							cptr = cptrkeep;
						}
						else
						{
							fdlist_del_event(cptr->fdlistinfo.fdlist,
								cptr->fdlistinfo.fdlist_pos,
								FDLIST_FLAGS_WRITE);
							continue;
						}
					}
					else
					{
						if (!cptr->queue_out.queue_size) /* should be false */
							fdlist_del_event(cptr->fdlistinfo.fdlist,
								cptr->fdlistinfo.fdlist_pos,
								FDLIST_FLAGS_WRITE);
					}
				}
				if ((cptr->maxsendq != -1) &&
						(cptr->queue_out.queue_size > cptr->maxsendq))
				{
					sprintf(exitbuf, "Max SendQ Exceeded: %i > %i",
							cptr->queue_out.queue_size, cptr->maxsendq);
					client_del(meptr, cptr, exitbuf);
					continue;
				}
				if (client_queue_flush(cptr, exitbuf))
					client_del(meptr, cptr, exitbuf);
			}
		}
		break;
	}
	return 0;
}

static void flood_check(void)
{
	IRCD_CLIENT	*fptr;
	IRCD_CLIENT	*nextf;

	for(fptr=Flooders;fptr;fptr=nextf)
	{
		nextf = fptr->f_next;
		if (Now > fptr->rate_time)
		{
			ircd_parse(fptr);
			/* FIX...can't guarantee ->f_next is valid */
		}
	}
}

static void ping_check(void)
{
	IRCD_CLIENT	*pptr;
	IRCD_CLIENT	*nextlo;

	for(pptr=Locals;pptr;pptr=nextlo)
	{
		nextlo = pptr->lo_next;

		if (IsDead(pptr))
		{
			if (pptr->auth_ptr && (pptr->auth_ptr == nextlo))
				nextlo = nextlo->lo_next;
			client_del(meptr, pptr, "Dead client");
			continue;
		}

		if (IsDoingDns(pptr) && (Now <= pptr->firsttime + 30))
		{
			if (myres_retry_request(pptr->res_stuff->res_status.rptr))
			{
				myres_status_deinit(&(pptr->res_stuff->res_status));
				pptr->flags &= ~CLIENT_FLAGS_DOINGDNS;
			}
			else
			{
				continue;
			}
		}
		if (IsUnknown(pptr) && (Now > pptr->firsttime + 30))
		{
			if (IsDoingConnect(pptr) || (!IsDoingDns(pptr) && !IsDoingAuth(pptr)))
			{
				if (pptr->auth_ptr && (pptr->auth_ptr == nextlo)) /* shouldnt*/
					nextlo = nextlo->lo_next;
				client_del(meptr, pptr, "Too much time as unknown");
				continue;
			}
			if (IsDoingAuth(pptr))
			{
				if (pptr->auth_ptr) /* assume this should be true! */
				{
					client_del(meptr, pptr->auth_ptr, "Auth timeout.");
				}
			}
			if (IsDoingDns(pptr))
			{
				myres_status_deinit(&(pptr->res_stuff->res_status));
				pptr->flags &= ~CLIENT_FLAGS_DOINGDNS;
			}
			pptr->firsttime = Now;
			ircd_parse(pptr);
			/*
			** hate to start loop over, but I have to :(
			**  we can't guarantee that ->next is valid
			*/
			nextlo = Locals;
			continue;
		}
		if (pptr->pingfreq && ((Now - pptr->lasttime) > pptr->pingfreq))
		{
			if (pptr->flags & CLIENT_FLAGS_PINGSENT)
			{
				if ((Now - pptr->lasttime) > (pptr->pingfreq*2))
				{
					char buffer[128];

					sprintf(buffer,
						"Ping timeout: no data for %i seconds",
						Now - pptr->lasttime);

					if (pptr->auth_ptr && (pptr->auth_ptr == nextlo))
						nextlo = nextlo->lo_next;
					
					client_del(meptr, pptr, buffer);
					continue;
				}
			}
			else
			{
				send_to_client(pptr, "PING :%s", meptr->name);
				pptr->flags |= CLIENT_FLAGS_PINGSENT;
			}
		}
		/* fdlist check */
		if (IsClient(pptr))
		{
			int		fdlistnum = Now - pptr->idletime;
			FDLIST	*fdlist = Fdlist_main;
			int		mfd;
			int		ret;
			int		flags;
			int		nfdlists;
			int		ptime;

			if (!IsAnOper(pptr) && ((nfdlists = (int)vars_value_get_bynum(IRCD_VAR_PRIORITY_FDLISTS)) > 0))
			{
				ptime = (int)vars_value_get_bynum(IRCD_VAR_PRIORITY_IDLE);
				fdlistnum /= ptime;
				if (fdlistnum > 0)
				{
					fdlistnum--;
					if (fdlistnum >= nfdlists)
						fdlistnum = nfdlists - 1;
					fdlist = &Fdlist_clients[fdlistnum];
				}
			}
			if (pptr->fdlistinfo.fdlist != fdlist)
			{
				flags = fdlist_get_events(pptr->fdlistinfo.fdlist,
					pptr->fdlistinfo.fdlist_pos);
				assert(flags);
				ret = fdlist_del_fd(pptr->fdlistinfo.fdlist,
                        pptr->fdlistinfo.fdlist_pos, &mfd);
				if (mfd >= 0)
				{
					if (GET_FDS(mfd))
						GET_FDS(mfd)->fdlistinfo.fdlist_pos =
	                        pptr->fdlistinfo.fdlist_pos;
				}
				ret = ircd_open_fd(pptr->fdlistinfo.fd, flags, pptr, fdlist);
				if (ret < 0)
				{
					if (pptr->auth_ptr && (pptr->auth_ptr == nextlo))
						nextlo = nextlo->lo_next;
					client_del(meptr, pptr, "Fdlist problem");
				}
			}
		}
	}
}

static void servlink_check(void)
{
	SERVLINK	*sptr;

	for(sptr=_ircd_conf.servers;sptr;sptr=sptr->next)
	{
		if ((sptr->flags & SERVLINK_FLAGS_DOINGDNS) &&
						myres_retry_request(sptr->res_status.rptr))
		{
			myres_status_deinit(&(sptr->res_status));
			sptr->flags &= ~SERVLINK_FLAGS_DOINGDNS;
			send_to_opers("Lookup of %s for \"link %s\" conf entry failed",
					sptr->hostname, sptr->servername);
		}
	}
}

static void kline_check(void)
{
	IRCD_CLIENT		*pptr;
	IRCD_CLIENT		*nextlo;
	int				ret;
	char			exitbuf[512];
	CONFHOSTIPENTRY	*entryptr;
	char			cnamebuf[CLIENTNAMESIZE];

	for(pptr=Locals;pptr;pptr=nextlo)
	{
		nextlo = pptr->lo_next;

		if (!IsClient(pptr))
			continue;

		if (pptr->class->flags & CLASS_FLAGS_BYPASSBANS)
			continue;

		ret = access_match_ipaddr(&(_ircd_conf.deny),
				pptr->username, pptr->ipstring,
				pptr->rem_ipaddr, ACCESS_FLAGS_BESTMATCH,
				&entryptr);
		if (ret < 0)
			ret = access_match_hostname(&(_ircd_conf.deny),
				pptr->username, pptr->hostname, ACCESS_FLAGS_BESTMATCH,
				&entryptr);

		if (ret >= 0) /* banned */
		{
			send_to_opers("Kill line active for %s",
					ircd_clientname(pptr, 0, cnamebuf));
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_YOUREBANNEDCREEP, pptr->name),
					entryptr->data ? entryptr->data : "No reason specified");
			sprintf(exitbuf, "Banned: %s",
				entryptr->data ? entryptr->data : "No reason specified");
			if (pptr->auth_ptr && (pptr->auth_ptr == nextlo))
				nextlo = nextlo->lo_next;
			client_del(meptr, pptr, exitbuf);
			continue;
		}
	}
}

static void rejecthold_check(void)
{
	IRCD_CLIENT	*rptr;
	IRCD_CLIENT	*nextrptr;
	int			num = (int)vars_value_get_bynum(IRCD_VAR_REJECT_HOLD);

	for(rptr=Rejectholds;rptr;rptr=rptr=nextrptr)
	{
		nextrptr = rptr->lo_next;
		if ((num <= 0) || ((rptr->firsttime+num) <= Now))
		{
			client_del(meptr, rptr, "bye now");
		}
	}
}

static int autoconn_check(void)
{
	SERVLINK		*s;
	SERVLINK		*s2;
	IRCD_CLIENT		*sptr;
	IRCD_CLASSENTRY	*classptr;
	int				ret;
	int				nextcheck = 300;
	int				diff;

	for(s=_ircd_conf.servers;s;s=s->next)
	{
		if (!(s->flags & SERVLINK_FLAGS_AUTOCONN))
			continue;
		if (Now < s->tryconnecttime + 60)
		{
			continue;
		}
		sptr = hash_find_client_incl_hm(s->servername);
		if (!sptr)
		{
			/*
			** ok, this one doesn't exist...let's see if another one in
			**	the same class does exist...
			*/
			if (s->classptr)
				classptr = (IRCD_CLASSENTRY *)s->classptr;
			else
			{
				ret = class_find(&(_ircd_conf.classes), s->classname, &classptr);

				if (ret)
					continue;
				s->classptr = (void *)classptr;
			}
			for(sptr=classptr->clients;sptr;sptr=sptr->cl_next)
				if (IsServer(sptr))
					break;
			if (sptr)		/* yup!, don't autoconn this one */
				continue;

			if ((diff = classptr->lastconnect + classptr->connfreq) <= Now)
			{
				int ret2;

				for(s2=_ircd_conf.servers;s2;s2=s2->next)
				{
					if (!s2->classptr)
					{
						IRCD_CLASSENTRY *classptr2;

						ret = class_find(&(_ircd_conf.classes),
										s2->classname, &classptr2);
						if (ret)
							continue;
						s2->classptr = (void *)classptr2;
					}
					if (s2 == s)
						continue;
					if ((s2->classptr == s->classptr) &&
						(Now < s2->tryconnecttime + 60))
							break;
				}
				if (s2)
					continue; /* already trying to connect to another */
				ret2 = _do_connect(NULL, s, -1);

				classptr->lastconnect = Now;
				if (classptr->connfreq < nextcheck)
					nextcheck = classptr->connfreq;

				if (!ret2)
				{
					servlink_rotate(s);
					break;
				}
			}
			else if (diff - Now < nextcheck)
				nextcheck = diff - Now;
		}
		else
		{
			if (s->classptr)
				classptr = (IRCD_CLASSENTRY *)s->classptr;
			else
			{
				ret = class_find(&(_ircd_conf.classes), s->classname, &classptr);

				if (ret)
					continue;
				s->classptr = (void *)classptr;
			}
			if (classptr->connfreq < nextcheck)
				nextcheck = classptr->connfreq;
		}
	}
	return nextcheck < 10 ? 10 : nextcheck;
}

static int _ircdso_firsttime(int argc, char **argv)
{
	int				ret;
	register int	i;
	int				nofork = 0;
	int				opt;
	int				optlen;

	if (argc < 2)
	{
		fprintf(stderr, "Usage: ircd ircd.conf\n");
		exit(-1);
	}
	
	if ((argc > 2) && (*argv[1] == '-'))
	{
		if (argv[1][1] == 'n')
		{
			nofork = 1;
			argc--;
			argv++;
		}
	}

	Ircd_zero_fd = open("/dev/zero", O_RDWR);
	if (Ircd_zero_fd < 0)
	{
		fprintf(stderr, "Couldn't open /dev/zero\n");
		exit(-1);
	}

	if ((i = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		fprintf(stderr, "couldn't socket(): %i\n", -errno);
		exit(-1);
	}

	optlen = sizeof(opt);
	if (getsockopt(i, SOL_SOCKET, SO_RCVBUF, (char *)&opt, &optlen) < 0)
	{
		fprintf(stderr, "couldn't getsockopt(): %i\n", -errno);
		exit(-1);
	}

	while(setsockopt(i, SOL_SOCKET, SO_RCVBUF, (char *)&opt, sizeof(opt)) >= 0)
	{
		opt += 1024;
		if (opt > (64*1024))
			break;
	}

	optlen = sizeof(opt);
	if (getsockopt(i, SOL_SOCKET, SO_RCVBUF, (char *)&opt, &optlen) < 0)
	{
		fprintf(stderr, "couldn't getsockopt(): %i\n", -errno);
		exit(-1);
	}

	Recvbuffersize = opt;

	optlen = sizeof(opt);
	if (getsockopt(i, SOL_SOCKET, SO_SNDBUF, (char *)&opt, &optlen) < 0)
	{
		fprintf(stderr, "couldn't getsockopt(): %i\n", -errno);
		exit(-1);
	}

	while(setsockopt(i, SOL_SOCKET, SO_SNDBUF, (char *)&opt, sizeof(opt)) >= 0)
	{
		opt += 1024;
		if (opt > (64*1024))
			break;
	}

	optlen = sizeof(opt);
	if (getsockopt(i, SOL_SOCKET, SO_SNDBUF, (char *)&opt, &optlen) < 0)
	{
		fprintf(stderr, "couldn't getsockopt(): %i\n", -errno);
		exit(-1);
	}

	Sendbuffersize = opt;
	close(i);

#ifdef MALLOC_DEBUG
	dmalloc_init(32*1024-1,32*1024-1,8);
#endif
	Conffile = MyStrdup(argv[1]);

	meptr = (IRCD_CLIENT *)MyMalloc(sizeof(IRCD_CLIENT));
	Fds = (IRCD_CLIENT **)MyMalloc(sizeof(IRCD_CLIENT *) * Maxconnections);
	Sentalong = (int *)MyMalloc(sizeof(int) * Maxconnections);
	Sentalong_marker = 0;
	Sentalong_end = Sentalong + Maxconnections;
	Globalinfo = (GLOBALSTATS *)MyMalloc(sizeof(GLOBALSTATS));
	Globalinfo->m_maxconntime = Globalinfo->g_maxclientstime = 
			Globalinfo->g_maxserverstime = Globalinfo->maxchannelstime =
			Globalinfo->m_maxclientstime = Globalinfo->m_maxserverstime =
			Globalinfo->m_totclientstime = (int)time(NULL);
	Globalinfo->g_servers = 1;

	Fdlist_main = (FDLIST *)MyMalloc(sizeof(FDLIST));
	Fdlist_clients = (FDLIST *)MyMalloc(sizeof(FDLIST) * Maxfdlists);
	Fdlist_lasttime = (int *)MyMalloc(sizeof(int) * Maxfdlists);

	Ircd_resfd = myres_init();
	if (Ircd_resfd < 0)
	{
		fprintf(stderr, "Couldn't do myres_init: %i\n", Ircd_resfd);
		exit(-1);
	}

	Ircd_resclient = (IRCD_CLIENT *)MyMalloc(sizeof(IRCD_CLIENT));
	Ircd_resclient->type = CLIENT_TYPE_RES;

#if 0
	ircd_pagesize = ircd_getpagesize();
#else
	ircd_pagesize = 2*1024;
#endif

	vars_init();
	numeric_init();
	ircdconf_init(&_ircd_conf);
	fdlist_init(Fdlist_main);

	for(i=0;i<Maxfdlists;i++)
	{
		fdlist_init(&Fdlist_clients[i]);
		Fdlist_lasttime[i] = 0;
	}

	iphash_init();

	if (ircd_open_fd(Ircd_resfd, FDLIST_FLAGS_READ, Ircd_resclient, NULL) < 0)
	{
		fprintf(stderr, "Error putting resfd into fdlist\n");
		exit(-1);
	}

	Now = (int)time(NULL);
	ret = ircdconf_read(Conffile, &_ircd_conf);
	if (ret < 0)
	{
		fprintf(stderr, "Couldn't read config file \"%s\": %i\n",
					Conffile, ret);
		exit(ret);
	}
	if (!_ircd_conf.name)
	{
		fprintf(stderr, "No server name given in ircd.conf\n");
		exit(-1);
	}

	{
		char    *motd_file = vars_value_get_bynum(IRCD_VAR_MOTD_FILE);

		if (motd_file)
			motd_load(motd_file, &Motd);
	}

	whowas_init();
	strncpy(meptr->name, _ircd_conf.name, sizeof(meptr->name)-1);
	strncpy(meptr->ircname, _ircd_conf.info ? _ircd_conf.info : "Unknown",
			sizeof(meptr->ircname)-1);
	meptr->ircname[sizeof(meptr->ircname)-1] = '\0';
	meptr->type = CLIENT_TYPE_SERVER;
	meptr->from = meptr;
	meptr->servptr = meptr;
	meptr->firsttime = meptr->lasttime = Now;
	hash_add_client(meptr);
	add_client_to_oslist(&Servers, meptr);

	if (!_ircd_conf.listeners)
	{
		fprintf(stderr, "No ports on which to listen\n");
		exit(-1);
	}

	fclose(stdin);
	fclose(stdout);

#ifdef TIOCNOTTY
	{
		int fd = open("/dev/tty", O_RDWR);

		if (fd >= 0)
		{
			ioctl(fd, TIOCNOTTY, NULL);
			close(fd);
		}
	}
#endif

	if (!nofork)
	{
		closelog();
		ret = fork();
		if (ret)
		{
			exit(0);
		}
		openlog("ircd-comstud", LOG_PID|LOG_NDELAY, LOG_DAEMON);
	}
	logs_init();
	syslog(LOG_NOTICE, "ircd has started.");

	{
		int fd = open("ircd.pid", O_CREAT|O_TRUNC|O_WRONLY, 0660);

		if (fd >= 0)
		{
			char buf[10];

			sprintf(buf, "%i\n", (int)getpid());
			write(fd, buf, strlen(buf));
			close(fd);
		}
	}
	return 0;
}

static int _ircdso_init(int firsttime)
{
	int			needconversion;
	IRCD_DATA	*Coreptr = *((IRCD_DATA **)ircdcore_data);

	needconversion = 0;

	signal(SIGHUP, _ircd_sighup_handler);
	signal(SIGUSR1, _ircd_sigusr1_handler);

	command_tree_init();
	hash_init();

	if (firsttime)
	{
		return _ircdso_firsttime(ircdcore_argc, ircdcore_argv);
	}
	else if (needconversion)
	{
		;
	}
	else
	{
		if (!Coreptr || (*ircdcore_data_size != sizeof(IRCD_DATA)))
			return -EFAULT;
		Maxconnections = Coreptr->Maxconnections;
		Maxfdlists = Coreptr->Maxfdlists;
		Conffile = Coreptr->Conffile;
		ircd_pagesize = Coreptr->ircd_pagesize;
		Fdlist_main = Coreptr->Fdlist_main;
		Fdlist_clients = Coreptr->Fdlist_clients;
		Fdlist_lasttime = Coreptr->Fdlist_lasttime;
		meptr = Coreptr->meptr;
		Fds = Coreptr->Fds;
		Locals = Coreptr->Locals;
		Clients = Coreptr->Clients;
		Opers = Coreptr->Opers;
		Servers = Coreptr->Servers;
		Flooders = Coreptr->Flooders;
		Channels = Coreptr->Channels;
		Globalinfo = Coreptr->Globalinfo;
		Rejectholds = Coreptr->Rejectholds;
		Motd = Coreptr->Motd;
		Sendbuffersize = Coreptr->Sendbuffersize;
		Recvbuffersize = Coreptr->Recvbuffersize;
		Sentalong = Coreptr->Sentalong;
		Sentalong_marker = Coreptr->Sentalong_marker;
		Sentalong_end = Coreptr->Sentalong_end;
		Ircd_resfd = Coreptr->Ircd_resfd;
		Ircd_zero_fd = Coreptr->Ircd_zero_fd;
		BCOPY(&(Coreptr->Ircd_conf), &_ircd_conf, sizeof(_ircd_conf));
		queue_set_stats(Coreptr->queue_size, Coreptr->queue_malloced_size);
		myres_init_reload(Coreptr->Ircd_resfd, Coreptr->res_requests, Coreptr->res_lastrequest);
		numeric_init_reload(Coreptr->NumericArr, Coreptr->NumericArrSize);
		whowas_init_reload(Coreptr->whowasarr, Coreptr->whowas_size,
			Coreptr->whowashash, Coreptr->whowas_hash_size,
			Coreptr->whowas_next);
		vars_init_reload(Coreptr->Vars);
		hash_init_reload(Coreptr->client_hash, Coreptr->client_hash_size,
			Coreptr->channel_hash, Coreptr->channel_hash_size);
		iphash_init_reload(Coreptr->iphash, Coreptr->iphash_size,
				Coreptr->clones, Coreptr->throttles);
	}

	DoRehash = 0;
	RehashWho = NULL;
	Rehashed = 0;
	Autoconncheck = 0;

	return 0;
}

static int _ircdso_deinit()
{
	IRCD_DATA	*Coreptr = *((IRCD_DATA **)ircdcore_data);


	if (*ircdcore_data)
		MyFree(*ircdcore_data);
	*ircdcore_data = (char *)MyMalloc(sizeof(IRCD_DATA));
	Coreptr = *((IRCD_DATA **)ircdcore_data);

	*ircdcore_data_size = sizeof(IRCD_DATA);
	*ircdcore_data_format_version = 0;

#ifdef HAVE_POLL
	if (pfdarray)
		MyFree(pfdarray);
#endif

	Coreptr->Maxconnections = Maxconnections;
	Coreptr->Conffile = Conffile;
	Coreptr->ircd_pagesize = ircd_pagesize;
	Coreptr->Fdlist_main = Fdlist_main;
	Coreptr->Fdlist_clients = Fdlist_clients;
	Coreptr->Fdlist_lasttime = Fdlist_lasttime;
	Coreptr->meptr = meptr;
	Coreptr->Fds = Fds;
	Coreptr->Locals = Locals;
	Coreptr->Clients = Clients;
	Coreptr->Opers = Opers;
	Coreptr->Servers = Servers;
	Coreptr->Flooders = Flooders;
	Coreptr->Channels = Channels;
	Coreptr->Globalinfo = Globalinfo;
	Coreptr->Rejectholds = Rejectholds;
	Coreptr->Motd = Motd;
	Coreptr->Sendbuffersize = Sendbuffersize;
	Coreptr->Recvbuffersize = Recvbuffersize;
	Coreptr->Sentalong = Sentalong;
	Coreptr->Sentalong_marker = Sentalong_marker;
	Coreptr->Sentalong_end = Sentalong_end;
	Coreptr->Ircd_resfd = Ircd_resfd;
	Coreptr->Ircd_zero_fd = Ircd_zero_fd;
	queue_get_stats(&Coreptr->queue_size, NULL, &Coreptr->queue_malloced_size);
	myres_deinit_reload(&Coreptr->res_requests, &Coreptr->res_lastrequest);
	BCOPY(&_ircd_conf, &(Coreptr->Ircd_conf), sizeof(_ircd_conf));
	numeric_deinit_reload(&Coreptr->NumericArr, &Coreptr->NumericArrSize);
	whowas_deinit_reload(&Coreptr->whowasarr, &Coreptr->whowas_size,
		&Coreptr->whowashash, &Coreptr->whowas_hash_size,
		&Coreptr->whowas_next);
	vars_deinit_reload(&Coreptr->Vars);
	hash_deinit_reload(&Coreptr->client_hash, &Coreptr->client_hash_size,
		&Coreptr->channel_hash, &Coreptr->channel_hash_size);
	iphash_deinit_reload(&Coreptr->iphash, &Coreptr->iphash_size,
			&Coreptr->clones, &Coreptr->throttles);

	if (RehashWho)
		MyFree(RehashWho);

	command_tree_free();

	signal(SIGHUP, SIG_IGN);
	signal(SIGUSR1, SIG_IGN);

	closelog();

	return 0;
}


int _ircd_main_loop(int argc,
					char **argv,
					int firsttime,
					char *Ircdcore_versionstring,
					int	*Data_format_version,
					char **Data,
					int *Data_size)
{
	int			ret;
	FDLIST		*curfdlist;
	int			i;

	if (!Data_format_version || !Data || !Data_size ||
			!Ircdcore_versionstring)
		return -EINVAL;

	ircdcore_argc = argc;
	ircdcore_argv = argv;
	ircdcore_versionstring = Ircdcore_versionstring;
	ircdcore_data_format_version = Data_format_version;
	ircdcore_data = Data;
	ircdcore_data_size = Data_size;

	if (firsttime)
	{
		Maxconnections = (*((IRCD_DATA **)Data))->Maxconnections;
		Maxfdlists = (*((IRCD_DATA **)Data))->Maxfdlists;
		if ((Maxconnections <= 0) || (Maxfdlists <= 0))
			return -EINVAL;
	}
	else
		openlog("ircd-comstud", LOG_PID|LOG_NDELAY, LOG_DAEMON);

	if (*ircdcore_data_format_version > 0)
		return -EINVAL;

	ret = _ircdso_init(firsttime);
	if (ret < 0)
		return ret;

	if (!firsttime)
		send_to_opers("ircd.so has been reloaded");

#ifdef HAVE_POLL
	pfdarray = (FDLIST_POLLFD *)MyMalloc(Maxconnections * sizeof(FDLIST_POLLFD));
#endif

	curfdlist = Fdlist_main;
	for(;;)
	{
		if (DoRehash)
		{
			ircdconf_rehash();
			DoRehash = 0;
			Autoconncheck = 0;
		}

		if (DoReload)
		{
			ret = _ircdso_deinit();

			if (ret < 0)
			{
				send_to_opers("Couldn't prepare for ircd.so unload: %i", ret);
			}
			else
			{
				return 0;
			}
		}

		ret = _ircd_pollfds(curfdlist, 2);

		if (Now >= (Fdlist_lasttime[0] + (int)vars_value_get_bynum(IRCD_VAR_PRIORITY_LAG)))
		{
			_ircd_pollfds(&Fdlist_clients[0], 0);
			Fdlist_lasttime[0] = Now;
			for(i=1;i<(int)vars_value_get_bynum(IRCD_VAR_PRIORITY_FDLISTS);i++)
			{
				if (Now >= (Fdlist_lasttime[i] + ((i+1)*(int)vars_value_get_bynum(IRCD_VAR_PRIORITY_LAG))))
				{
					_ircd_pollfds(&Fdlist_clients[i], 0);
					Fdlist_lasttime[i] = Now;
				}
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 2))
			{
				flood_check();
				lastcheck = Now;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 5))
			{
				ping_check();
				lastcheck = Now;
			}
		}

		if (Now >= (Globalinfo->bw_last_check + 5))
		{
			if (Globalinfo->bw_last_check)
			{
				Globalinfo->bw_c_incoming_rate = Globalinfo->bw_c_incoming_bytes/(Now - Globalinfo->bw_last_check);
				Globalinfo->bw_c_outgoing_rate = Globalinfo->bw_c_outgoing_bytes/(Now - Globalinfo->bw_last_check);
				Globalinfo->bw_s_incoming_rate = Globalinfo->bw_s_incoming_bytes/(Now - Globalinfo->bw_last_check);
				Globalinfo->bw_s_outgoing_rate = Globalinfo->bw_s_outgoing_bytes/(Now - Globalinfo->bw_last_check);
			}
			Globalinfo->bw_c_incoming_bytes = 0;
			Globalinfo->bw_c_outgoing_bytes = 0;
			Globalinfo->bw_s_incoming_bytes = 0;
			Globalinfo->bw_s_outgoing_bytes = 0;
			Globalinfo->bw_last_check = Now;
		}

		{
			static int	lastcheck = 0;

			if (Rehashed || (Now >= (lastcheck + 900)))
			{
				kline_check();
				lastcheck = Now;
				Rehashed = 0;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 60))
			{
				access_flush(&_ircd_conf.deny, 0);
				lastcheck = Now;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 5))
			{
				servlink_check();
				lastcheck = Now;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 10))
			{
				iphash_cleanup();
				lastcheck = Now;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 900))
			{
				queue_cleanup();
				lastcheck = Now;
			}	
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 10))
			{
				rejecthold_check();
				lastcheck = Now;
			}
		}

		if (Now >= Autoconncheck)
		{
			Autoconncheck = Now + autoconn_check();
		}
	}

	return 0;
}
